import { getHighlighter } from './shiki/shiki';
import { cache, generateHash, codeLanguage, renderToHtml } from './utils';
import { selectAll } from 'unist-util-select';
import memoize from 'micro-memoize';
import pAll from 'p-all';

// @ts-ignore
const getHighlightLineNumbers = (str: string): number[] | undefined => {
  if (!str) return;
  let numbers: number[] | undefined = undefined;
  numbers = str.split(',').flatMap(s => {
    if (!s.includes('-')) return +s;

    const [min, max] = s.split('-');
    // @ts-ignore
    const final = Array.from({ length: max - min + 1 }, (_, n) => n + +min);
    return final;
  });
  return numbers;
};

const getProps = (meta: any) => {
  let props =
    meta &&
    meta.split(' ').reduce((acc: any, cur: any) => {
      if (cur && cur.includes('=')) {
        const t = cur.split('=');
        acc[t[0]] = t[1];
        return acc;
      }

      acc[cur] = true;
      return acc;
    }, {});
  if (props.highlight) {
    props.highlight = getHighlightLineNumbers(props.highlight);
  }
  return props;
};

const DEFAULT_THEME = 'Material-Theme-Darker-High-Contrast';

const generateHighlighter = memoize(
  async () =>
    getHighlighter({
      theme: DEFAULT_THEME,
      renderer: renderToHtml,
      cache,
    }),
  { isPromise: true }
);

// @ts-ignore
const highlight = (options: any) => async (node: any) => {
  const props = node && node.meta && getProps(node.meta);
  const highlighter = await generateHighlighter();
  // @ts-ignore
  return highlighter.codeToHtml(node, props);
};

const tryCachedNode = (options: any) => async (node: any) => {
  const hash = generateHash(node.value);
  const cached = await cache.get(hash);
  if (cached) {
    node.value = cached;
  } else {
    const lang = codeLanguage(node);
    node.lang = lang;
    const highlighted = await highlight(options)(node);
    await cache.set(hash, highlighted);
    node.properties = node.meta;
  }
};

const visitor = (options: any) => async (node: any) => {
  node.type = 'html';
  node.children = undefined;
  if (process.env.NODE_ENV === 'production') {
    // we cache nodes because they get highlighted multiple times during build
    // and we don't need to redo it each time
    await tryCachedNode(options)(node);
  } else {
    const lang = codeLanguage(node);
    node.lang = lang;
    node.value = await highlight(options)(node);
  }
};

const transformer = (options: any) =>
  memoize(
    async (tree: any) => {
      const nodes = selectAll('code', tree);
      await pAll(
        nodes.map(node => () => visitor(options)(node)),
        { concurrency: 25 }
      );
      return tree;
    },
    { isPromise: true }
  );

function attacher(options: any) {
  return transformer(options);
}

module.exports = memoize(attacher);
